<!--use <pre> tag for displaying content-->
<pre>{{content}}</pre>

<!--make a caption to the content -->
{% if cp.check_option(filename, 'caption') %}
    <p><i>{{cp.get(filename, 'caption')}}</i></p>
{% endif %}

<!--always link to file-->
<a class="btn btn-xs btn-primary" href='{{filename}}' target="_blank">Link to File</a>

<!--show cmd line if there is one -->
{% if cp.check_option(filename, 'cmd') %}
    <button type="button" class="btn btn-xs btn-primary"
            data-toggle="popover" title="command line"
            data-content="{{cp.get(filename, 'cmd')}}">
            Command Line
    </button>
{% endif %}
