

                {% if cp.check_option(filename, 'title') %}
                    <a href='#'>{{cp.get(filename, 'title')}}</a>
                {% else %}
                    <a href='#'>{{filename}}</a>
                {% endif %}

                <pre>{{content}}</pre>

                <!--make a caption to the content -->
                {% if cp.check_option(filename, 'caption') %}
                    <p><i>{{cp.get(filename, 'caption')}}</i></p>
                {% endif %}            
            
                <!--always link to file-->
                <p><a href={{filename}}>Link to file.</a></p>
            
                <!--show cmd line if there is one -->
                {% if cp.check_option(filename, 'cmd') %}
                    <button type="button" class="btn btn-xs"
                            data-toggle="popover" title="command line"
                            data-content="{{cp.get(filename, 'cmd')}}">
                            Command Line
                    </button>
                {% endif %}
