# coding: utf-8
# Generated by setup.py for PyCBC on 2021-02-23 10:17:40 +0000.

version = '1.18.0'
date = '2021-02-23 09:53:18 +0000'
release = True
last_release = '1.18.0'

git_hash = 'daf32ffcc068dba03673bc6dd6b3fc1c515f60a8'
git_branch = 'None'
git_tag = 'v1.18.0'
git_author = 'Alex Nitz <alex.nitz@gmail.com>'
git_committer = 'GitHub <noreply@github.com>'
git_status = 'CLEAN: All modifications committed'
git_builder = 'Unknown User <>'
git_build_date = '2021-02-23 10:17:40 +0000'
git_verbose_msg = """Version: 1.18.0
Branch: None
Tag: v1.18.0
Id: daf32ffcc068dba03673bc6dd6b3fc1c515f60a8
Builder: Unknown User <>
Build date: 2021-02-23 10:17:40 +0000
Repository status is CLEAN: All modifications committed"""
from pycbc._version import *
