# Exactly 18706 templates expected

pycbc_geom_aligned_bank --pn-order threePointFivePN --f0 60 --f-low 60 --f-upper 1000 --delta-f 0.01 --min-match 0.97 --min-mass1 1. --min-mass2 1. --max-mass1 5. --max-mass2 5. --min-total-mass 2.5 --max-total-mass 6.0 --min-chirp-mass 1.2187707886145738 --max-chirp-mass 2.4375415772291475 --min-eta 0.16 --max-eta 0.24 --verbose --log-path /usr1/spxiwh/logs --max-ns-spin-mag 0.5 --max-bh-spin-mag 0.9 --ns-bh-boundary-mass 2.0 --output-file testAligned.xml --split-bank-num 50 --asd-file ZERO_DET_high_P.txt  --random-seed 20 --calculate-time-metric-components --filter-cutoff SchwarzISCO --filter-points
